"""Essential tests for CLI functionality."""

import pytest
from promptheus.cli import build_parser, parse_arguments


def test_build_parser_basic():
    """Test that the argument parser builds correctly."""
    parser = build_parser(include_subcommands=False)
    
    # Should have the basic arguments
    assert parser is not None


def test_parse_args_history_command():
    """Test parsing with history subcommand."""
    parser = build_parser(include_subcommands=True)
    args = parser.parse_args(["history"])
    assert args.command == "history"
    assert not args.clear
    assert args.limit == 20


def test_parse_args_prompt_value_not_treated_as_command():
    """Ensure free-form prompt strings are accepted when no subcommand is used."""
    args = parse_arguments(["Write a blog post"])

    assert args.prompt == "Write a blog post"
    assert args.command is None


def test_parse_args_prompt_allows_options_after_text():
    """Prompt tokens followed by flags should still parse correctly."""
    args = parse_arguments(["Write a blog post", "--copy"])

    assert args.prompt == "Write a blog post"
    assert args.copy is True


def test_parse_args_history_clear():
    """Test parsing with history clear."""
    parser = build_parser(include_subcommands=True)
    args = parser.parse_args(["history", "--clear"])
    assert args.command == "history"
    assert args.clear


def test_parse_args_history_limit():
    """Test parsing with history limit."""
    parser = build_parser(include_subcommands=True)
    args = parser.parse_args(["history", "--limit", "50"])
    assert args.command == "history"
    assert args.limit == 50


def test_parse_args_file_flag():
    """Test parsing with file flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--file", "test.txt"])
    assert args.file == "test.txt"


def test_parse_args_with_provider():
    """Test parsing with provider flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--provider", "gemini"])
    assert args.provider == "gemini"


def test_parse_args_with_model():
    """Test parsing with model flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--model", "gpt-4"])
    assert args.model == "gpt-4"


def test_parse_args_with_flags():
    """Test parsing with various boolean flags."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--skip-questions", "--copy"])
    assert args.skip_questions
    assert args.copy


def test_parse_args_skip_questions_flag():
    """Test parsing with skip-questions flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--skip-questions"])
    assert args.skip_questions

    # Test the short form
    args = parser.parse_args(["-s"])
    assert args.skip_questions


def test_parse_args_refine_flag():
    """Test parsing with refine flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--refine"])
    assert args.refine


def test_parse_args_verbose_flag():
    """Test parsing with verbose flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["-v"])
    assert args.verbose


def test_parse_args_file_and_provider():
    """Test that file and provider arguments work together correctly."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["--file", "test.txt", "--provider", "openai"])
    assert args.file == "test.txt"
    assert args.provider == "openai"


def test_parse_args_multiple_flags():
    """Test parsing with multiple flags."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args([
        "--provider", "anthropic",
        "--model", "claude-3",
        "--skip-questions",
        "--copy"
    ])
    assert args.provider == "anthropic"
    assert args.model == "claude-3"
    assert args.skip_questions
    assert args.copy


def test_parse_args_output_format():
    """Test parsing with output format flag."""
    parser = build_parser(include_subcommands=False)
    args = parser.parse_args(["-o", "json"])
    assert args.output_format == "json"

    args = parser.parse_args(["--output-format", "plain"])
    assert args.output_format == "plain"


def test_build_parser_help():
    """Test that help can be generated without errors."""
    parser = build_parser(include_subcommands=True)
    
    # This should not raise an exception
    help_text = parser.format_help()
    assert "Promptheus" in help_text
    assert "AI-powered prompt engineering CLI tool" in help_text



"""Question prompter abstraction for interactive and non-interactive modes."""

from __future__ import annotations

import sys
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

import questionary

from promptheus.io_context import IOContext, MessageSink


class QuestionPrompter(ABC):
    """Abstract interface for prompting users with questions."""

    @abstractmethod
    def prompt_confirmation(self, message: str, default: bool = True) -> bool:
        """
        Ask a yes/no question.

        Args:
            message: The question to ask
            default: Default answer if user just presses enter

        Returns:
            True for yes, False for no

        Raises:
            EOFError: If stdin is not interactive
        """
        pass

    @abstractmethod
    def prompt_radio(self, message: str, choices: List[str]) -> str:
        """
        Ask user to select one option from a list.

        Args:
            message: The question to ask
            choices: List of options to choose from

        Returns:
            Selected option

        Raises:
            EOFError: If stdin is not interactive
        """
        pass

    @abstractmethod
    def prompt_checkbox(self, message: str, choices: List[str]) -> List[str]:
        """
        Ask user to select multiple options from a list.

        Args:
            message: The question to ask
            choices: List of options to choose from

        Returns:
            List of selected options

        Raises:
            EOFError: If stdin is not interactive
        """
        pass

    @abstractmethod
    def prompt_text(self, message: str, default: str = "") -> str:
        """
        Ask user for text input.

        Args:
            message: The question to ask
            default: Default value if user just presses enter

        Returns:
            User's text input

        Raises:
            EOFError: If stdin is not interactive
        """
        pass


class RichPrompter(QuestionPrompter):
    """
    Interactive prompter using questionary for rich TTY experience.

    Uses arrow keys, colored output, and interactive menus.
    Requires both stdin and stdout to be TTYs.
    """

    def prompt_confirmation(self, message: str, default: bool = True) -> bool:
        answer = questionary.confirm(message, default=default).ask()
        if answer is None:
            raise KeyboardInterrupt("User cancelled")
        return answer

    def prompt_radio(self, message: str, choices: List[str]) -> str:
        answer = questionary.select(message, choices=choices).ask()
        if answer is None:
            raise KeyboardInterrupt("User cancelled")
        return answer

    def prompt_checkbox(self, message: str, choices: List[str]) -> List[str]:
        answer = questionary.checkbox(message, choices=choices).ask()
        if answer is None:
            raise KeyboardInterrupt("User cancelled")
        return answer

    def prompt_text(self, message: str, default: str = "") -> str:
        answer = questionary.text(message, default=default).ask()
        if answer is None:
            raise KeyboardInterrupt("User cancelled")
        return answer.strip()


class StdioPrompter(QuestionPrompter):
    """
    Simple prompter using stderr/input() for when stdout is piped.

    All prompts written to stderr, reads from stdin.
    Works when stdin is TTY but stdout is not.
    """

    def __init__(self, notify: MessageSink):
        """
        Initialize stdio prompter.

        Args:
            notify: Function to write messages to stderr
        """
        self.notify = notify

    def prompt_confirmation(self, message: str, default: bool = True) -> bool:
        default_str = "Y/n" if default else "y/N"
        sys.stderr.write(f"\n{message} ({default_str}): ")
        sys.stderr.flush()
        try:
            response = input().strip().lower()
        except EOFError:
            raise EOFError("stdin is not interactive")

        if not response:
            return default
        return response in ('y', 'yes')

    def prompt_radio(self, message: str, choices: List[str]) -> str:
        sys.stderr.write(f"\n{message}\n")
        for i, choice in enumerate(choices, 1):
            sys.stderr.write(f"  {i}. {choice}\n")
        sys.stderr.write(f"Choice [1-{len(choices)}]: ")
        sys.stderr.flush()

        while True:
            try:
                user_input = input().strip()
            except EOFError:
                raise EOFError("stdin is not interactive")

            try:
                idx = int(user_input) - 1
                if 0 <= idx < len(choices):
                    return choices[idx]
                else:
                    sys.stderr.write(f"Invalid choice. Please enter 1-{len(choices)}: ")
                    sys.stderr.flush()
            except ValueError:
                sys.stderr.write("Invalid input. Please enter a number: ")
                sys.stderr.flush()

    def prompt_checkbox(self, message: str, choices: List[str]) -> List[str]:
        sys.stderr.write(f"\n{message} (comma-separated numbers)\n")
        for i, choice in enumerate(choices, 1):
            sys.stderr.write(f"  {i}. {choice}\n")
        sys.stderr.write(f"Choices [1-{len(choices)}]: ")
        sys.stderr.flush()

        while True:
            try:
                user_input = input().strip()
            except EOFError:
                raise EOFError("stdin is not interactive")

            if not user_input:
                return []

            try:
                indices = [int(c.strip()) - 1 for c in user_input.split(',') if c.strip()]
                selected = [choices[i] for i in indices if 0 <= i < len(choices)]
                if selected or not user_input:
                    return selected
                else:
                    sys.stderr.write("Invalid choices. Please try again: ")
                    sys.stderr.flush()
            except ValueError:
                sys.stderr.write("Invalid input. Please enter comma-separated numbers: ")
                sys.stderr.flush()

    def prompt_text(self, message: str, default: str = "") -> str:
        default_str = f" [{default}]" if default else ""
        sys.stderr.write(f"\n{message}{default_str}: ")
        sys.stderr.flush()

        try:
            answer = input().strip()
        except EOFError:
            raise EOFError("stdin is not interactive")

        if not answer and default:
            return str(default)
        return answer


def create_prompter(io: IOContext) -> QuestionPrompter:
    """
    Factory function to create appropriate prompter based on I/O context.

    Args:
        io: I/O context with TTY state

    Returns:
        RichPrompter if fully interactive, StdioPrompter otherwise
    """
    if io.is_fully_interactive:
        return RichPrompter()
    else:
        return StdioPrompter(io.notify)



"""Essential tests for provider functionality."""

import json
import os
from unittest.mock import Mock, patch, MagicMock
import pytest

from promptheus.providers import (
    LLMProvider,
    GeminiProvider,
    AnthropicProvider,
    OpenAIProvider,
    GroqProvider,
    QwenProvider,
    GLMProvider,
    get_provider,
    _parse_question_payload,
    _append_json_instruction,
    _build_chat_messages,
    _coerce_message_content
)
from promptheus.config import Config


class MockProvider(LLMProvider):
    """Mock provider for testing purposes."""
    
    def generate_questions(self, initial_prompt: str, system_instruction: str):
        # Not used in these tests but required for abstract base class
        return None
    
    def get_available_models(self):
        return ["test-model"]
    
    def _generate_text(self, prompt: str, system_instruction: str, json_mode: bool = False, max_tokens=None):
        return "Mocked response from provider"


def test_abstract_base_class():
    """Test that LLMProvider can't be instantiated without implementing abstract methods."""
    with pytest.raises(TypeError, match="abstract methods"):
        LLMProvider()


def test_parse_question_payload_valid():
    """Test parsing of valid JSON question payload."""
    valid_json = '{"task_type": "generation", "questions": [{"question": "Test?"}]}'
    result = _parse_question_payload("TestProvider", valid_json)
    
    assert result == {"task_type": "generation", "questions": [{"question": "Test?"}]}


def test_parse_question_payload_invalid_json():
    """Test parsing of invalid JSON."""
    invalid_json = "not json at all"
    result = _parse_question_payload("TestProvider", invalid_json)
    
    assert result is None


def test_parse_question_payload_missing_task_type():
    """Test parsing of JSON without task_type."""
    json_without_task_type = '{"questions": [{"question": "Test?"}]}'
    result = _parse_question_payload("TestProvider", json_without_task_type)
    
    assert result is None


def test_parse_question_payload_adds_empty_questions():
    """Test that missing questions field gets added."""
    json_without_questions = '{"task_type": "analysis"}'
    result = _parse_question_payload("TestProvider", json_without_questions)
    
    assert result == {"task_type": "analysis", "questions": []}


def test_append_json_instruction():
    """Test appending JSON instruction to prompts."""
    result = _append_json_instruction("Hello world")
    expected_suffix = (
        "Respond ONLY with a valid JSON object using double-quoted keys. "
        "Include the fields specified in the instructions (for example, task_type and questions). "
        "Do not wrap the JSON in markdown code fences or add commentary."
    )
    
    assert result.endswith(expected_suffix)


def test_append_json_instruction_idempotent():
    """Test that appending JSON instruction is idempotent."""
    prompt = "Hello world"
    first_result = _append_json_instruction(prompt)
    second_result = _append_json_instruction(first_result)
    
    assert first_result == second_result


def test_build_chat_messages():
    """Test building chat messages from system instruction and prompt."""
    messages = _build_chat_messages("system", "user prompt")
    
    assert len(messages) == 2
    assert messages[0] == {"role": "system", "content": "system"}
    assert messages[1] == {"role": "user", "content": "user prompt"}


def test_build_chat_messages_no_system():
    """Test building chat messages with no system instruction."""
    messages = _build_chat_messages("", "user prompt")
    
    assert len(messages) == 1
    assert messages[0] == {"role": "user", "content": "user prompt"}


def test_coerce_message_content_string():
    """Test coercing message content from string."""
    result = _coerce_message_content("hello world")
    assert result == "hello world"


def test_coerce_message_content_list():
    """Test coercing message content from list."""
    content_list = [{"text": "hello"}, {"content": "world"}]
    result = _coerce_message_content(content_list)
    assert result == "helloworld"


def test_coerce_message_content_with_attr():
    """Test coercing message content from object with text attribute."""
    mock_obj = Mock()
    mock_obj.text = "hello from object"
    result = _coerce_message_content(mock_obj)
    assert result == "hello from object"


def test_provider_factory_gemini(monkeypatch):
    """Test provider factory function for Gemini."""
    monkeypatch.setenv("GEMINI_API_KEY", "test-key")

    sample_models = {
        "providers": {
            "gemini": {
                "default_model": "gemini-pro",
                "models": ["gemini-pro", "gemini-1.5-pro"],
                "api_key_env": "GEMINI_API_KEY",
                "api_key_prefixes": ["AIza"],
            }
        },
        "provider_aliases": {"gemini": "Gemini"},
    }
    monkeypatch.setattr("promptheus.config.Config.load_provider_config", lambda self: sample_models)
    
    config = Config()
    config.set_provider("gemini")
    
    provider = get_provider("gemini", config, "gemini-pro")
    
    assert isinstance(provider, GeminiProvider)
    assert provider.model_name == "gemini-pro"


def test_provider_factory_openai(monkeypatch):
    """Test provider factory function for OpenAI."""
    monkeypatch.setenv("OPENAI_API_KEY", "test-key")
    
    sample_models = {
        "providers": {
            "openai": {
                "default_model": "gpt-4o",
                "models": ["gpt-4o", "gpt-4o-mini"],
                "api_key_env": "OPENAI_API_KEY",
                "api_key_prefixes": ["sk-"],
            }
        },
        "provider_aliases": {"openai": "OpenAI"},
    }
    monkeypatch.setattr("promptheus.config.Config.load_provider_config", lambda self: sample_models)
    
    config = Config()
    config.set_provider("openai")
    
    provider = get_provider("openai", config, "gpt-4o-mini")
    
    assert isinstance(provider, OpenAIProvider)
    assert provider.model_name == "gpt-4o-mini"


def test_provider_factory_unknown():
    """Test provider factory function with unknown provider."""
    config = Mock()
    
    with pytest.raises(ValueError, match="Unknown provider: unknown"):
        get_provider("unknown", config)


def test_format_refinement_payload():
    """Test formatting of refinement payload."""
    # Since _format_refinement_payload is a concrete method that doesn't use abstract methods,
    # we can test it by creating a mock provider that implements abstract methods
    provider = MockProvider()
    
    payload = provider._format_refinement_payload(
        "Initial prompt",
        {"audience": "developers", "tone": "professional"},
        {"audience": "Who is the audience?", "tone": "What tone to use?"}
    )
    
    assert "Initial Prompt: Initial prompt" in payload
    assert "Who is the audience?: developers" in payload
    assert "What tone to use?: professional" in payload
    assert "Please generate a refined, optimized prompt" in payload


def test_format_tweak_payload():
    """Test formatting of tweak payload."""
    # Use the mock provider to test the concrete method
    provider = MockProvider()
    
    payload = provider._format_tweak_payload(
        "Current prompt",
        "Make it more formal"
    )
    
    assert "Current Prompt:" in payload
    assert "Current prompt" in payload
    assert "User's Modification Request:" in payload
    assert "Make it more formal" in payload
    assert "Return the tweaked prompt:" in payload


def test_refine_from_answers():
    """Test refining prompt from answers."""
    provider = MockProvider()
    # Mock the _generate_text method to return a specific value
    with patch.object(provider, '_generate_text', return_value="refined prompt") as mock_gen:
        result = provider.refine_from_answers(
            "initial",
            {"audience": "developers"},
            {"audience": "Who is the audience?"},
            "system instruction"
        )
        
        assert result == "refined prompt"
        mock_gen.assert_called_once()
        # Check that the call includes the formatted payload
        call_args = mock_gen.call_args
        assert "initial" in call_args[0][0]  # first argument is the prompt
        assert "Who is the audience?: developers" in call_args[0][0]


def test_tweak_prompt():
    """Test tweaking prompt with instruction."""
    provider = MockProvider()
    with patch.object(provider, '_generate_text', return_value="tweaked prompt") as mock_gen:
        result = provider.tweak_prompt(
            "current prompt",
            "Make it shorter",
            "system instruction"
        )
        
        assert result == "tweaked prompt"
        mock_gen.assert_called_once()
        call_args = mock_gen.call_args
        assert "current prompt" in call_args[0][0]
        assert "Make it shorter" in call_args[0][0]


def test_light_refine():
    """Test light refinement functionality."""
    provider = MockProvider()
    with patch.object(provider, '_generate_text', return_value="light refined prompt") as mock_gen:
        result = provider.light_refine("initial prompt", "system instruction")

        assert result == "light refined prompt"
        mock_gen.assert_called_once()


# ========================
# OpenAI-Compatible Provider Utility Tests
# ========================

def test_coerce_message_content_various_types():
    """Test _coerce_message_content handles various input types."""
    from promptheus.providers import _coerce_message_content

    # Test empty list
    result = _coerce_message_content([])
    assert result == ""

    # Test mixed types in list
    content_list = [{"text": "hello"}, "world", {"type": "text", "content": "test"}]
    result = _coerce_message_content(content_list)
    assert result == "hellotest"

    # Test None input
    result = _coerce_message_content(None)
    assert result == ""

    # Test numeric input
    result = _coerce_message_content(123)
    assert result == "123"




"""Tests for output format functionality."""

import pytest
from promptheus.cli import parse_arguments


def test_output_format_flag_parsing():
    """Test that -o/--output-format flag is parsed correctly."""
    # Test plain (default)
    args = parse_arguments(["test prompt"])
    assert args.output_format == "plain"

    # Test plain explicitly
    args = parse_arguments(["-o", "plain", "test prompt"])
    assert args.output_format == "plain"

    # Test json
    args = parse_arguments(["--output-format", "json", "test prompt"])
    assert args.output_format == "json"


def test_output_format_default_value():
    """Test that output format has correct default."""
    args = parse_arguments(["test prompt"])
    assert args.output_format == "plain"



"""
Shared provider utilities used by completion and validation commands.

This module contains common provider-related functionality that is used
across different command modules to avoid code duplication.
"""

import os
from typing import Optional, Tuple

from promptheus.config import Config
from promptheus.providers import get_provider
from promptheus.utils import sanitize_error_message


def _select_test_model(provider_name: str, config: Config) -> Optional[str]:
    """Pick a provider-specific model for connection testing."""
    provider_meta = config._ensure_provider_config().get("providers", {}).get(provider_name, {})

    # 1) Provider-specific MODEL env vars (e.g., OPENAI_MODEL) take precedence
    model_env = provider_meta.get("model_env")
    if model_env:
        env_value = os.getenv(model_env)
        if env_value:
            return env_value

    # 2) Use the provider's curated default model from providers.json
    default_model = provider_meta.get("default_model")
    if default_model:
        return default_model

    # 3) Fall back to the first example model if provided
    examples = provider_meta.get("example_models") or []
    if examples:
        return examples[0]

    # 4) Absolute fallback: global override (PROMPTHEUS_MODEL)
    return os.getenv("PROMPTHEUS_MODEL")


def _test_provider_connection(provider_name: str, config: Config) -> Tuple[bool, str]:
    """Attempt a simple API call to test credentials for a provider."""
    original_provider = None
    try:
        # Temporarily set the provider to ensure we use its settings for the test
        original_provider = config.provider
        config.set_provider(provider_name)

        test_model = _select_test_model(provider_name, config)
        provider = get_provider(provider_name, config, model_name=test_model)

        # Use a simple, low-cost prompt for testing
        provider._generate_text("ping", "", max_tokens=8)
        return True, ""
    except Exception as exc:
        if provider_name == "gemini" and "did not include text content" in str(exc):
            return False, "Connection failed. Note: Standard Gemini API keys may have limitations."
        return False, sanitize_error_message(str(exc))
    finally:
        # Restore original provider setting
        if original_provider:
            config.set_provider(original_provider)
        else:
            config.reset() # Clear the temporary provider setting


def _filter_text_models(models, excludes=None):
    """Filter out models that are not text-based."""
    if excludes is None:
        excludes = ("embed", "embedding", "image", "vision", "audio", "speech", "video", "sound", "draw", "paint", "whisper")

    filtered = []
    for model in models:
        lower = model.lower()
        if any(token in lower for token in excludes):
            continue
        filtered.append(model)
    return filtered


#!/usr/bin/env python3
"""
Promptheus - AI-powered prompt engineering CLI tool.
Unified version with multi-provider support.
"""

from __future__ import annotations

import logging
import os
import sys
from argparse import Namespace
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Tuple

import questionary
import pyperclip
from rich.console import Console
from rich.panel import Panel
from rich.text import Text

from promptheus.config import Config
from promptheus.constants import PROMPTHEUS_DEBUG_ENV, VERSION
from promptheus.history import get_history
from promptheus.io_context import IOContext
from promptheus.question_prompter import create_prompter
from promptheus.prompts import (
    ANALYSIS_REFINEMENT_SYSTEM_INSTRUCTION,
    CLARIFICATION_SYSTEM_INSTRUCTION,
    GENERATION_SYSTEM_INSTRUCTION,
    TWEAK_SYSTEM_INSTRUCTION,
)
from promptheus.providers import LLMProvider, get_provider
from promptheus.utils import configure_logging, sanitize_error_message
from promptheus.cli import parse_arguments
from promptheus.repl import display_history, interactive_mode
from promptheus.exceptions import PromptCancelled, ProviderAPIError, InvalidProviderError
from promptheus.commands import list_models, validate_environment, generate_template, generate_completion_script, handle_completion_request

console = Console()
logger = logging.getLogger(__name__)

MessageSink = Callable[[str], None]


@dataclass
class QuestionPlan:
    skip_questions: bool
    task_type: str
    questions: List[Dict[str, Any]]
    mapping: Dict[str, str]
    use_light_refinement: bool = False


def convert_json_to_question_definitions(
    questions_json: List[Dict[str, Any]]
) -> Tuple[List[Dict[str, Any]], Dict[str, str]]:
    """
    Convert JSON question format to internal question definitions.
    Handles required vs optional questions with visual indicators.
    """
    question_defs: List[Dict[str, Any]] = []
    question_mapping: Dict[str, str] = {}

    for idx, q in enumerate(questions_json):
        question_text = q.get("question", f"Question {idx + 1}")
        question_type = q.get("type", "text")
        options = q.get("options", [])
        required = q.get("required", True)

        key = f"q{idx}"
        question_mapping[key] = question_text

        display_text = question_text if required else f"{question_text} (optional)"

        question_defs.append(
            {
                "key": key,
                "message": display_text,
                "type": question_type.lower(),
                "options": options or [],
                "required": required,
                "default": q.get("default", ""),
            }
        )

    return question_defs, question_mapping


def display_output(
    prompt: str,
    io: IOContext,
    is_refined: bool = True,
) -> None:
    """Display the prompt in a panel or as plain text."""
    if io.quiet_output:
        # In quiet mode, don't display anything here - output is handled by the caller
        return

    # In non-quiet mode, display the Rich panel on stderr
    title = "[bold green]Refined Prompt[/bold green]" if is_refined else "[bold blue]Your Prompt[/bold blue]"
    border_color = "green" if is_refined else "blue"

    prompt_text = Text(prompt)
    panel = Panel(prompt_text, title=title, border_style=border_color, padding=(1, 2))

    io.console_err.print("\n")
    io.console_err.print(panel)
    io.console_err.print()


def copy_to_clipboard(text: str, notify: Optional[MessageSink] = None) -> None:
    """Copy text to clipboard."""
    if notify is None:
        notify = console.print
    try:
        pyperclip.copy(text)
        notify("[green]✓[/green] Copied to clipboard!")
    except Exception as exc:  # pragma: no cover - platform dependent
        sanitized = sanitize_error_message(str(exc))
        notify(f"[yellow]Warning: Failed to copy to clipboard: {sanitized}[/yellow]")
        logger.exception("Clipboard copy failed")


def iterative_refinement(
    provider: LLMProvider,
    current_prompt: str,
    io: IOContext,
    plain_mode: bool,
) -> str:
    """
    Allow user to iteratively refine the prompt with simple tweaks.
    Returns the final accepted prompt.
    """
    # Skip iterative refinement in quiet mode
    if io.quiet_output:
        return current_prompt

    iteration = 1

    while True:
        try:
            if plain_mode:
                tweak_instruction = input("Tweak? (Enter to accept, or describe your change) ").strip()
            else:
                answer = questionary.text(
                    "Tweak? (Enter to accept, or describe your change)"
                ).ask()
                if answer is None:
                    io.notify("\n[yellow]Cancelled tweaks.[/yellow]\n")
                    return current_prompt
                tweak_instruction = answer.strip()
        except (EOFError, KeyboardInterrupt):
            io.notify("\n[yellow]Cancelled tweaks.[/yellow]\n")
            return current_prompt

        if not tweak_instruction:
            io.notify("\n[green]✓[/green] Prompt accepted!\n")
            return current_prompt

        iteration += 1

        io.notify(f"\n[blue]⟳[/blue] Tweaking prompt (v{iteration})...\n")

        try:
            with io.console_err.status("[bold cyan]✨ Sprinkling some refinement magic...", spinner="bouncingBall"):
                current_prompt = provider.tweak_prompt(
                    current_prompt, tweak_instruction, TWEAK_SYSTEM_INSTRUCTION
                )

            display_output(current_prompt, io, is_refined=True)

        except KeyboardInterrupt:
            io.notify("\n[yellow]Cancelled tweaks.[/yellow]\n")
            return current_prompt
        except Exception as exc:
            sanitized = sanitize_error_message(str(exc))
            io.notify(f"[bold red]Error:[/bold red] Failed to tweak prompt: {sanitized}")
            io.notify("[yellow]Keeping previous version[/yellow]\n")
            logger.exception("Prompt tweak failed")


def determine_question_plan(
    provider: LLMProvider,
    initial_prompt: str,
    args: Namespace,
    debug_enabled: bool,
    io: IOContext,
    app_config: Config,
) -> QuestionPlan:
    """
    Decide whether to ask clarifying questions and prepare them if needed.
    """
    if getattr(args, "skip_questions", False):
        io.notify("\n[bold blue]✓[/bold blue] Skip questions mode - improving prompt directly\n")
        return QuestionPlan(skip_questions=True, task_type="analysis", questions=[], mapping={})

    try:
        if not io.quiet_output:
            with io.console_err.status("[bold magenta]🔍 Analyzing your prompt and crafting questions...", spinner="arc"):
                result = provider.generate_questions(initial_prompt, CLARIFICATION_SYSTEM_INSTRUCTION)
        else:
            result = provider.generate_questions(initial_prompt, CLARIFICATION_SYSTEM_INSTRUCTION)
    except ProviderAPIError as exc:
        current_provider = app_config.provider or ""
        provider_display = current_provider.title() if current_provider else "Provider"
        # Pass a larger max_length to prevent truncation of the core error message
        sanitized_error = sanitize_error_message(str(exc), max_length=500)

        io.notify(f"\n[bold red]✗ {provider_display} Error:[/bold red] [red]{sanitized_error}[/red]")
        io.notify(f"[dim](set {PROMPTHEUS_DEBUG_ENV}=1 to print full debug output)[/dim]\n")

        available_providers = app_config.get_configured_providers()
        other_providers = [p for p in available_providers if p != current_provider]
        provider_label = current_provider or "default"

        if other_providers:
            io.notify(f"[dim]💡 Current provider: '[cyan]{provider_label}[/cyan]'. Perhaps try a different one?[/dim]")
            for p in other_providers:
                io.notify(f"[dim]  - [cyan]promptheus --provider {p} ...[/cyan][/dim]")
        else:
            io.notify("[dim]💡 Double-check your credentials, or use '--quick' for offline questions.[/dim]")
        io.notify("")
        raise RuntimeError("AI provider unavailable for question generation") from exc
    except KeyboardInterrupt as exc:
        raise PromptCancelled("Analysis cancelled") from exc

    if result is None:
        # This block should now be much harder to reach, but serves as a fallback.
        current_provider = app_config.provider or ""
        provider_display = current_provider.title() if current_provider else "Provider"
        io.notify(
            f"\n[bold yellow]⚠ {provider_display} is taking a break![/bold yellow] "
            f"[dim](set {PROMPTHEUS_DEBUG_ENV}=1 to print debug output)[/dim]"
        )
        io.notify(f"[dim]Reason: Your {provider_display} provider returned an unexpected empty result.[/dim]")
        io.notify("[dim]We need a working AI to generate questions for your prompt.[/dim]")
        io.notify("")
        raise RuntimeError("AI provider returned an unexpected empty result.")

    task_type = result.get("task_type", "generation")
    questions_json = result.get("questions", [])

    if debug_enabled:
        io.notify(
            f"[dim]Debug: task_type={task_type}, questions={len(questions_json)}, refine={args.refine}[/dim]"
        )

    if task_type == "analysis" and not args.refine:
        io.notify("\n[bold blue]✓[/bold blue] Analysis task detected - performing light refinement")
        io.notify("[dim]  (Use --skip-questions to skip, or --refine to force questions)[/dim]\n")
        return QuestionPlan(True, task_type, [], {})

    if not questions_json:
        io.notify("\n[bold blue]✓[/bold blue] No clarifying questions needed\n")
        return QuestionPlan(True, task_type, [], {})

    if task_type == "generation" and not args.refine:
        io.notify(
            f"\n[bold green]✓[/bold green] Creative task detected with {len(questions_json)} clarifying questions"
        )
        try:
            # Create appropriate prompter based on I/O context
            prompter = create_prompter(io)
            confirm = prompter.prompt_confirmation(
                "Ask clarifying questions to refine your prompt?", default=True
            )
        except EOFError:
            io.notify("[yellow]stdin not interactive - skipping questions, using light refinement[/yellow]")
            return QuestionPlan(True, task_type, [], {}, use_light_refinement=True)
        except KeyboardInterrupt:
            io.notify("[yellow]Skipping questions - performing light refinement[/yellow]")
            return QuestionPlan(True, task_type, [], {}, use_light_refinement=True)
        if not confirm:
            io.notify("\n[bold]Skipping questions - performing light refinement\n")
            return QuestionPlan(True, task_type, [], {}, use_light_refinement=True)

    questions, mapping = convert_json_to_question_definitions(questions_json)
    if args.refine:
        io.notify(f"[bold green]✓[/bold green] Refine mode - {len(questions)} questions generated\n")
    return QuestionPlan(False, task_type, questions, mapping)


def ask_clarifying_questions(
    plan: QuestionPlan,
    io: IOContext,
) -> Optional[Dict[str, Any]]:
    """Prompt the user with clarifying questions and return their answers."""
    if plan.skip_questions or not plan.questions:
        return {}

    io.notify("[bold]Please answer the following questions to refine your prompt:[/bold]\n")

    answers: Dict[str, Any] = {}

    # Create appropriate prompter based on I/O context
    prompter = create_prompter(io)

    for question in plan.questions:
        key = question["key"]
        qtype = question.get("type", "text").lower()
        message = question.get("message", key)
        required = question.get("required", True)
        options = question.get("options") or []
        default = question.get("default", "")

        while True:
            try:
                if qtype == "radio" and options:
                    answer = prompter.prompt_radio(message, options)
                elif qtype == "checkbox" and options:
                    answer = prompter.prompt_checkbox(message, options)
                elif qtype == "confirm":
                    default_bool = bool(default) if isinstance(default, bool) else True
                    answer = prompter.prompt_confirmation(message, default_bool)
                else:
                    answer = prompter.prompt_text(message, str(default))
            except EOFError:
                io.notify("[red]Error: stdin is not interactive but questions need to be answered[/red]")
                io.notify("[dim]Use --skip-questions to skip questions or provide input via interactive terminal[/dim]")
                return None
            except KeyboardInterrupt:
                io.notify("[yellow]Cancelled.[/yellow]")
                return None

            # Normalize answer
            normalized = answer
            if isinstance(normalized, str):
                normalized = normalized.strip()

            if qtype == "checkbox":
                normalized = normalized or []

            missing_response = False
            if isinstance(normalized, str):
                missing_response = normalized == ""
            elif isinstance(normalized, list):
                missing_response = len(normalized) == 0

            if required and missing_response and qtype != "confirm":
                io.notify("[yellow]This answer is required. Please provide a response.[/yellow]")
                continue

            if not required and missing_response:
                normalized = [] if qtype == "checkbox" else ""

            answers[key] = normalized
            break

    return answers


def generate_final_prompt(
    provider: LLMProvider,
    initial_prompt: str,
    answers: Dict[str, Any],
    mapping: Dict[str, str],
    io: IOContext,
) -> Tuple[str, bool]:
    """Generate the refined prompt (or return original if no answers)."""
    if not answers:
        return initial_prompt, False

    try:
        if not io.quiet_output:
            with io.console_err.status("[bold green]🎨 Crafting your refined prompt...", spinner="moon"):
                final_prompt = provider.refine_from_answers(
                    initial_prompt, answers, mapping, GENERATION_SYSTEM_INSTRUCTION
                )
        else:
            final_prompt = provider.refine_from_answers(
                initial_prompt, answers, mapping, GENERATION_SYSTEM_INSTRUCTION
            )
        return final_prompt, True
    except KeyboardInterrupt as exc:
        raise PromptCancelled("Refinement cancelled") from exc
    except Exception as exc:
        sanitized = sanitize_error_message(str(exc))
        io.notify(f"[bold red]Error:[/bold red] Failed to generate refined prompt: {sanitized}")
        logger.exception("Failed to generate refined prompt")
        raise


def process_single_prompt(
    provider: LLMProvider,
    initial_prompt: str,
    args: Namespace,
    debug_enabled: bool,
    plain_mode: bool,
    io: IOContext,
    app_config: Config,
) -> Optional[Tuple[str, str]]:
    """
    Process a single prompt through the refinement pipeline.

    Returns:
        Tuple of (final_prompt, task_type) if successful, None otherwise
    """
    try:
        plan = determine_question_plan(provider, initial_prompt, args, debug_enabled, io, app_config)

        # This is the main logic branching
        is_light_refinement = (
            (plan.task_type == "analysis" and plan.skip_questions) or
            plan.use_light_refinement
        )

        if is_light_refinement:
            try:
                if not io.quiet_output:
                    with io.console_err.status("[bold blue]⚡ Performing light refinement...", spinner="simpleDots"):
                        final_prompt = provider.light_refine(
                            initial_prompt, ANALYSIS_REFINEMENT_SYSTEM_INSTRUCTION
                        )
                else:
                    final_prompt = provider.light_refine(
                        initial_prompt, ANALYSIS_REFINEMENT_SYSTEM_INSTRUCTION
                    )
                is_refined = True
            except KeyboardInterrupt as exc:
                raise PromptCancelled("Light refinement cancelled") from exc
            except Exception as exc:
                logger.warning("Light refinement failed: %s", sanitize_error_message(str(exc)))
                io.notify("[yellow]Warning: Light refinement failed. Using original prompt.[/yellow]")
                final_prompt = initial_prompt
                is_refined = False
        else:
            # The standard flow: ask questions if needed, then generate
            answers = ask_clarifying_questions(plan, io)
            if answers is None:
                return None
            final_prompt, is_refined = generate_final_prompt(
                provider, initial_prompt, answers, plan.mapping, io
            )

    except Exception as exc:
        sanitized = sanitize_error_message(str(exc))
        io.notify(f"[red]✗[/red] Something went wrong: {sanitized}")
        if not debug_enabled:
            io.notify(f"[dim]Enable --verbose for full error details[/dim]")
        logger.exception("Failed to process prompt")
        return None

    display_output(final_prompt, io, is_refined=is_refined)

    # Skip interactive tweaks in quiet mode or when skip-questions is set
    interactive_tweaks = io.stdin_is_tty and not getattr(args, "skip_questions", False) and not io.quiet_output
    if interactive_tweaks:
        final_prompt = iterative_refinement(provider, final_prompt, io, plain_mode)
    else:
        if getattr(args, "skip_questions", False):
            io.notify("[dim]Skipping interactive tweaking (skip-questions mode)[/dim]\n")
        elif io.quiet_output:
            # Don't notify in quiet mode
            pass
        else:
            io.notify("[dim]Skipping interactive tweaking (stdin is not a TTY)[/dim]\n")

    # Save to history
    try:
        history = get_history(app_config)
        history.save_entry(
            original_prompt=initial_prompt,
            refined_prompt=final_prompt,
            task_type=plan.task_type
        )
        logger.debug("Saved prompt to history")
    except Exception as exc:
        logger.warning(f"Failed to save prompt to history: {sanitize_error_message(str(exc))}")

    # Skip clipboard in quiet mode
    if not io.quiet_output:
        if getattr(args, "copy", False):
            copy_to_clipboard(final_prompt, io.notify)

    return final_prompt, plan.task_type


def main() -> None:
    """Main entry point for Promptheus."""
    configure_logging()
    app_config = Config()

    args = parse_arguments()

    # Handle version flag first (before any setup)
    if args.version:
        console.print(f"Promptheus v{VERSION}")
        sys.exit(0)

    # Create I/O context for terminal handling
    io = IOContext.create()

    if args.verbose:
        os.environ[PROMPTHEUS_DEBUG_ENV] = "1"
        configure_logging(logging.DEBUG)

    # Handle utility commands that exit immediately
    if getattr(args, "command", None) == "completion":
        if getattr(args, "install", False):
            from promptheus.commands import install_completion
            install_completion(args.shell, io.console_err)
        else:
            generate_completion_script(args.shell)
        sys.exit(0)

    if getattr(args, "command", None) == "__complete":
        # Note: __complete is a hidden command used by the shell completion scripts
        handle_completion_request(app_config, args)
        sys.exit(0)

    # Handle utility commands that exit immediately
    if getattr(args, "command", None) == "list-models":
        if args.providers:
            providers_to_list = [p.strip() for p in args.providers.split(',')]
        else:
            providers_to_list = None
        # Create a simple console for output
        utility_console = Console()
        list_models(
            app_config,
            utility_console,
            providers=providers_to_list,
            limit=args.limit,
            include_nontext=args.include_nontext
        )
        sys.exit(0)

    if getattr(args, "command", None) == "validate":
        providers_to_validate = None
        if args.providers:
            providers_to_validate = [p.strip() for p in args.providers.split(',')]

        # Create a simple console for output
        utility_console = Console()
        validate_environment(
            app_config,
            utility_console,
            test_connection=getattr(args, "test_connection", False),
            providers=providers_to_validate
        )
        sys.exit(0)

    if getattr(args, "command", None) == "template":
        # Create a simple console for output
        utility_console = Console()
        try:
            template = generate_template(app_config, utility_console, args.providers)
            utility_console.print(template)
        except Exception as exc:
            utility_console.print(f"[red]Error generating template:[/red] {exc}")
            sys.exit(1)
        sys.exit(0)

    if getattr(args, "command", None) == "history":
        if getattr(args, "clear", False):
            confirm = questionary.confirm(
                "Are you sure you want to clear all history?",
                default=False,
            ).ask()
            if confirm:
                get_history(app_config).clear()
                io.notify("[green]✓[/green] History cleared")
            else:
                io.notify("[yellow]Cancelled[/yellow]")
        else:
            display_history(io.console_err, io.notify, limit=args.limit)
        sys.exit(0)

    # Show provider status in a friendly way
    for message in app_config.consume_status_messages():
        io.notify(f"[cyan]●[/cyan] {message}")

    if args.provider:
        app_config.set_provider(args.provider)
    if args.model:
        app_config.set_model(args.model)

    for message in app_config.consume_status_messages():
        io.notify(f"[cyan]●[/cyan] {message}")

  
    # Friendly error handling
    if not app_config.validate():
        io.notify("")
        for message in app_config.consume_error_messages():
            # Split multi-line messages and format nicely
            lines = message.split('\n')
            if len(lines) == 1:
                io.notify(f"[red]✗[/red] {message}")
            else:
                io.notify(f"[red]✗[/red] {lines[0]}")
                for line in lines[1:]:
                    io.notify(f"  {line}")
        io.notify("")
        sys.exit(1)

    # Get initial prompt from file, stdin, or argument
    initial_prompt: Optional[str] = None

    if args.file:
        try:
            with open(args.file, "r", encoding="utf-8") as file_handle:
                initial_prompt = file_handle.read().strip()
            io.notify(f"[green]✓[/green] Loaded prompt from {args.file}")
        except FileNotFoundError:
            io.notify(f"[red]✗[/red] Couldn't find file: {args.file}")
            sys.exit(1)
        except Exception as exc:  # pragma: no cover - file I/O
            sanitized = sanitize_error_message(str(exc))
            io.notify(f"[red]✗[/red] Failed to read file: {sanitized}")
            sys.exit(1)

    elif args.prompt and args.prompt.startswith("@"):
        filename = args.prompt[1:]
        try:
            with open(filename, "r", encoding="utf-8") as file_handle:
                initial_prompt = file_handle.read().strip()
            io.notify(f"[green]✓[/green] Loaded prompt from {filename}")
        except FileNotFoundError:
            io.notify(f"[red]✗[/red] Couldn't find file: {filename}")
            sys.exit(1)
        except Exception as exc:  # pragma: no cover - file I/O
            sanitized = sanitize_error_message(str(exc))
            io.notify(f"[red]✗[/red] Failed to read file: {sanitized}")
            sys.exit(1)

    elif not io.stdin_is_tty:
        initial_prompt = sys.stdin.read().strip()
        if initial_prompt:
            io.notify("[green]✓[/green] Got prompt from stdin")

    else:
        initial_prompt = args.prompt

    provider_name = app_config.provider or "gemini"
    try:
        provider = get_provider(provider_name, app_config, app_config.get_model())
    except Exception as exc:
        error_msg = str(exc)
        sanitized = sanitize_error_message(error_msg)
        io.notify(f"[red]✗[/red] Couldn't connect to AI provider: {sanitized}\n")

        # Provide helpful context for common errors
        if "401" in error_msg or "403" in error_msg or "Unauthorized" in error_msg:
            io.notify(f"[yellow]Authentication Failed:[/yellow] Check your API key for {provider_name}\n")
        elif "404" in error_msg:
            io.notify(f"[yellow]Model Not Found:[/yellow] The model may not exist or be available\n")

        logger.exception("Provider initialization failure")
        sys.exit(1)

    debug_enabled = args.verbose or os.getenv(PROMPTHEUS_DEBUG_ENV, "").lower() in {"1", "true", "yes", "on"}
    plain_mode = io.plain_mode

    if initial_prompt is None or not initial_prompt:
        # Cannot enter interactive mode in quiet mode without a prompt
        if io.quiet_output:
            io.console_err.print("[red]✗[/red] Error: Cannot enter interactive mode when stdout is not a TTY")
            io.console_err.print("[dim]Provide a prompt as an argument, via --file, or from stdin[/dim]")
            io.console_err.print("[dim]Example: promptheus \"your prompt here\" | cat[/dim]")
            sys.exit(1)

        # Warn if history is disabled in interactive mode
        if not app_config.history_enabled:
            io.notify("[yellow]⚠[/yellow] History persistence is disabled. Your prompts won't be saved.\n")
            io.notify("[dim]Enable with: export PROMPTHEUS_ENABLE_HISTORY=1[/dim]\n")

        interactive_mode(
            provider,
            app_config,
            args,
            debug_enabled,
            plain_mode,
            io,
            process_single_prompt,
        )
    else:
        io.notify(f"[dim]Using provider: {provider_name} | Model: {app_config.get_model()}[/dim]\n")
        try:
            result = process_single_prompt(
                provider, initial_prompt, args, debug_enabled, plain_mode, io, app_config
            )
            if result:
                final_prompt, task_type = result
                # Write the final output to stdout based on output format
                output_format = getattr(args, "output_format", "plain")
                if output_format == "json":
                    import json
                    io.console_out.print(json.dumps({"prompt": final_prompt, "task_type": task_type}))
                else:  # plain
                    io.console_out.print(final_prompt)
            else:
                # process_single_prompt returned None - error occurred
                io.notify("[red]✗[/red] Failed to process prompt")
                sys.exit(1)
        except PromptCancelled as exc:
            io.notify(f"\n[yellow]{exc}[/yellow]\n")
            sys.exit(130)
        except KeyboardInterrupt:
            io.notify("\n[yellow]Cancelled by user[/yellow]\n")
            sys.exit(130)


if __name__ == "__main__":
    main()
